# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Script to compare the affected resources matched per-policy, per-region
between a dryrun and the last actual run of each policy. High-level results
are written to a Mardown file (to be added as a comment on the PR), Low-level
differences are written to a HTML file (to be added as a link on the PR).
"""

import sys
import glob
import re
import logging
import json
import boto3
import argparse
import itertools
import os
from zlib import decompress
import subprocess
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound

from c7n.resources import load_available
from c7n.provider import get_resource_class

from manheim_c7n_tools.utils import set_log_info, set_log_debug
from manheim_c7n_tools.config import ManheimConfig
from manheim_c7n_tools.version import VERSION

logger = logging.getLogger(__name__)


class DryRunDiffer(object):
    UNKNOWN_RESOURCE_TYPE = 'unknown_type'
    RESOURCE_TYPE_KEY = 'resource_type'
    UNKNOWN_RESOURCE_ID = 'unknown_id'

    def __init__(self, config):
        """
        Initialize a dryrun differ.

        :param config: manheim-c7n-tools configuration object
        :type config: ManheimConfig
        """
        self._live_results = {}
        self.config = config

    def run(self, git_dir=None, diff_against='master'):
        changed_policies = self._find_changed_policies(git_dir, diff_against)
        if len(changed_policies) == 0:
            logger.info(
                'Git diff did not report any changed policies; skipping '
                'resource count diff.'
            )
            with open('pr_diff.md', 'w') as fh:
                fh.write('Git diff did not report any changed policies; '
                         'skipping resource count diff.')
            return
        logger.info('Changed policies for diff: %s', changed_policies)
        dryrun_results = self._get_dryrun_results(changed_policies)
        logger.info('Reading results from last run from S3')
        for rname in self.config.regions:
            logger.debug('Getting S3 results for region: %s', rname)
            self._get_s3_results_for_region(rname, changed_policies)
        diff_md = self._make_diff_markdown(dryrun_results)
        with open('pr_diff.md', 'w') as fh:
            if 'defaults' in changed_policies:
                fh.write(
                    'PR found to contain changes to defaults.yml and '
                    '%d other policies\n\n' % (len(changed_policies) - 1)
                )
            else:
                fh.write(
                    'PR found to contain changes to '
                    '%d policies\n\n' % len(changed_policies)
                )
            fh.write(diff_md)
        logger.info('PR diff written to: pr_diff.md')
        diff_report = self._make_diff_report(dryrun_results)
        if diff_report:
            with open('pr_report.html', 'w') as fh:
                fh.write(diff_report)
            logger.info('PR report written to: pr_report.html')

    def _find_changed_policies(self, git_dir=None, diff_against='master'):
        """
        :return: list of policy names that differ from master
        :rtype: list
        """
        res = subprocess.check_output(
            ['git', 'diff', '--name-only', diff_against],
            cwd=git_dir
        ).decode().split("\n")
        pnames = []
        polname_re = re.compile(r'^policies.*/([a-zA-Z0-9_-]+)\.yml$')
        for x in res:
            x = x.strip()
            if x == '':
                continue
            m = polname_re.match(x)
            if not m:
                continue
            pnames.append(m.group(1))
        return pnames

    def _make_diff_report(self, dryrun):
        """
        Return a HTML report breaking down the differences between the dryrun
        (this branch) and the last run of each policy on master.

        Requires a jinja template located at `./reporting-template/report.j2`
        within the same directory where the dryrun-diff step has been run.

        :param dryrun: dryrun policy resource information
        :type dryrun: dict
        :return: html report
        :rtype: str
        """
        all_policies = list(
            set(dryrun.keys()) | set(self._live_results.keys())
        )
        entries = {}
        if not all_policies:
            logger.info('no policies found - skipping diff report')
            return ''
        t_loader = FileSystemLoader(searchpath="./reporting-template/")
        t_env = Environment(loader=t_loader)
        t_file = "report.j2"
        try:
            tmpl = t_env.get_template(t_file)
        except TemplateNotFound:
            logger.info('unable to find a template - skipping diff report')
            return ''

        for policy in all_policies:
            for region in self.config.regions:
                dryrun_id = self._get_resource_id(dryrun, policy)
                liverun_id = self._get_resource_id(self._live_results, policy)
                dryrun_res = dryrun.get(policy, {}).get(region, [])
                liverun_res = self._live_results.get(policy, {})\
                    .get(region, [])
                liverun_not_possible = \
                    liverun_res and liverun_id == self.UNKNOWN_RESOURCE_ID
                dryrun_not_possible = \
                    dryrun_res and dryrun_id == self.UNKNOWN_RESOURCE_ID
                if liverun_not_possible or dryrun_not_possible:
                    logger.error(
                        'unknown resource_id encountered for either \
                            dryrun or liverun resources - policy: %s', policy)
                    return ''
                dry_ids = {
                    '{}: {}'.format(dryrun_id, x[dryrun_id])
                    for x in dryrun_res
                }
                live_ids = {
                    '{}: {}'.format(liverun_id, x[liverun_id])
                    for x in liverun_res
                }
                additions = [
                    {'id': x, 'type': 'added'}
                    for x in dry_ids.difference(live_ids)
                ]
                removals = [
                    {'id': x, 'type': 'removed'}
                    for x in live_ids.difference(dry_ids)
                ]
                untouched = [
                    {'id': x, 'type': 'unchanged'}
                    for x in dry_ids.intersection(live_ids)
                ]
                resources = list(itertools.chain.from_iterable((
                    additions or [],
                    removals or [],
                    untouched or []
                )))
                if policy not in entries:
                    entries[policy] = {}
                entries[policy][region] = {
                    'resources': resources,
                    'total_add': len(additions),
                    'total_remove': len(removals),
                    'total_untouch': len(untouched)
                }
        return tmpl.render(
            account_name=self.config.account_id,
            entries=entries
        )

    def _get_resource_id(self, resource, policy):
        """
        Obtain the id for a given policy from a dict of resources.
        :param resource: the dict of resources
        :param policy: the name of the policy
        :return: the resource_id the policy affects
        :rtype: string
        """
        resource_type = resource.get(policy, {}).get(
            self.RESOURCE_TYPE_KEY, self.UNKNOWN_RESOURCE_TYPE)
        if resource_type == self.UNKNOWN_RESOURCE_TYPE:
            return
        load_available()
        _id = self.UNKNOWN_RESOURCE_ID
        try:
            _id = get_resource_class(resource_type) \
                .resource_type() \
                .id
        except Exception:
            logger.warning(
                'unable to get resource_id for %s - id: %s', policy, _id)
        return _id

    def _make_diff_markdown(self, dryrun):
        """
        Return GitHub-flavored Markdown showing the difference between the
        dryrun (this branch) and the last run of each policy on master.

        :param dryrun: dryrun policy resource counts
        :type dryrun: dict
        :return: markdown diff
        :rtype: str
        """
        all_policies = list(set(
            list(dryrun.keys()) + list(self._live_results.keys())
        ))
        if len(all_policies) == 0:
            return 'No policy reports found. Perhaps all changed policies ' \
                   'are event-based instead of periodic?\n'
        maxlen = max([len(x) for x in all_policies])
        fmt = '%s %{maxlen}s   %6s   %6s   %6s   \n'.format(maxlen=maxlen)
        linelen = 30 + maxlen
        res = '```diff\n'
        res += '##%s    last                  ##\n' % (' ' * maxlen)
        res += '##%s    live                  ##\n' % (' ' * maxlen)
        res += '##%s    run       PR    diff  ##\n' % (' ' * maxlen)
        for pname in sorted(all_policies):
            res += '%s\n' % ('=' * linelen)
            res += pname + "\n"
            for rname in self.config.regions:
                a = len(self._live_results.get(pname, {}).get(rname, []))
                b = len(dryrun.get(pname, {}).get(rname, []))
                a_str = '--' if a == 0 else a
                b_str = '--' if b == 0 else b
                prefix = ' '
                diff = ''
                if a == '--' and b != '--':
                    # in PR but not in master/live
                    prefix = '+'
                    diff = '+%d' % b
                elif a != '--' and b == '--':
                    # in master/live but not in PR
                    prefix = '-'
                    diff = -1 * a
                if a > b:
                    # in both, more resources in live than PR
                    prefix = '-'
                    diff = b - a
                elif a < b:
                    # in both, fewer resources in live than PR
                    prefix = '+'
                    diff = b - a
                res += fmt % (
                    prefix, rname, a_str, b_str, diff
                )
        res += '```\n'
        return res

    def _get_dryrun_results(self, pol_names):
        """
        Read the `resources.json` files from disk for the dryrun/ directory.
        Return a dictionary of string policy name to nested dictionaries, of
        string region name to resources.

        :return: dictionary of nested dictionaries, policy name to dict of
          region name to resource
        :rtype: dict
        """
        res = {}
        logger.debug('Getting dryrun results from disk...')
        fname_re = re.compile(r'dryrun/([^/]+)/([^/]+)/resources.json')
        for f in glob.glob('dryrun/*/*/resources.json'):
            m = fname_re.match(f)
            if not m:
                logger.error('ERROR: file path does not match regex: %s', f)
                continue
            region = m.group(1)
            policy = m.group(2)
            if policy not in pol_names and 'defaults' not in pol_names:
                # policy isn't changed
                continue
            _dir = os.path.dirname(f)
            logger.debug('Reading files from directory: %s', _dir)
            try:
                self._read_dryrun_files(_dir, policy, region, res)
            except Exception:
                logger.error('ERROR reading from dir: %s', _dir, exc_info=True)
                continue
        logger.debug('Got dryrun results for %d policies', len(res))
        return res

    def _read_dryrun_files(self, directory, pol, region, res):
        """
        Read the directory for dryrun files, and attaches each resource to the
        supplied result dict via their policy-name and region.
        :param directory: the path where the dryrun files are located
        :param pol: the name of the policy
        :param region: the name of the region
        :param res: the dict that will be mutated with the resources found.
        :return: a dict of resources
        :rtype: dict
        """
        if pol not in res:
            res[pol] = {}
        resources = os.path.join(directory, 'resources.json')
        with open(resources, 'r') as fh:
            logger.debug('Reading from file: %s', resources)
            res[pol][region] = json.loads(fh.read())
        if self.RESOURCE_TYPE_KEY not in res[pol]:
            metadata = os.path.join(directory, 'metadata.json')
            with open(metadata, 'r') as fh:
                logger.debug('Reading from file: %s', metadata)
                mj = json.loads(fh.read())
                _type = mj.get('policy', {}).get(
                    'resource', self.UNKNOWN_RESOURCE_TYPE)
                res[pol][self.RESOURCE_TYPE_KEY] = _type

    def _get_s3_results_for_region(self, region_name, changed_pols):
        """
        Find the results files in S3 from the last live run of the deployed
        policies. Reads each file and maps resources to ``self._live_results``
        accordingly.
        """
        s3 = boto3.resource('s3', region_name=region_name)
        bktname = self.config.config_for_region(
            region_name
        ).output_s3_bucket_name
        bkt = s3.Bucket(bktname)
        prefixes = self._get_s3_policy_prefixes(bkt)
        logger.debug('Found %d policy prefixes in %s', len(prefixes), bktname)
        for p in prefixes:
            if p not in changed_pols and 'defaults' not in changed_pols:
                # policy was not changed, skip it
                continue
            if p not in self._live_results:
                self._live_results[p] = {}
            fetch_type = self.RESOURCE_TYPE_KEY not in self._live_results[p]
            resource, _type = self._get_latest_res_for_policy(
                bkt, p, fetch_type)
            self._live_results[p][region_name] = resource
            if fetch_type:
                self._live_results[p][self.RESOURCE_TYPE_KEY] = _type
        logger.debug('Done getting resource counts for %s', region_name)

    def _get_s3_policy_prefixes(self, bucket):
        """
        Find all of the per-policy prefixes (a.k.a. "directories") in the S3
        bucket. Return a list of them

        :param bucket: the S3 bucket to list policies in
        :type bucket: ``boto3.S3.Bucket``
        :return: list of per-policy prefixes in S3 bucket
        :rtype: list
        """
        client = bucket.meta.client
        response = client.list_objects(
            Bucket=bucket.name,
            Delimiter='/',
            Prefix='logs/'
        )
        if response['IsTruncated']:
            raise RuntimeError('ERROR: S3 response was truncated!')
        if 'CommonPrefixes' not in response:
            logger.error(
                'ERROR: "CommonPrefixes" element not in S3 ListObjects '
                'response; bucket must be empty!'
            )
            return []
        result = []
        for pname in response['CommonPrefixes']:
            result.append(pname['Prefix'].replace('logs/', '').strip('/'))
        return result

    def _get_latest_res_for_policy(self, bucket, pol_name, get_res_type):
        """
        Given the S3 Bucket and a policy name, find the newest
        ``resources.json`` file for that policy and annotate it with it's
        respective type from the ``metadata.json`` file.

        :param bucket: the bucket to look in
        :type bucket: ``boto3.S3.Bucket``
        :param pol_name: the name of the policy
        :type pol_name: str
        :return: resource from latest run of the policy
        :rtype: obj
        """
        newest_res = None
        newest_meta = None
        for obj in bucket.objects.filter(Prefix='logs/%s/' % pol_name):
            if obj.key.endswith('/resources.json') or \
                    obj.key.endswith('/resources.json.gz'):
                if newest_res is None or \
                        obj.last_modified > newest_res.last_modified:
                    newest_res = obj
            if obj.key.endswith('/metadata.json') or \
                    obj.key.endswith('/metadata.json.gz'):
                if newest_meta is None or \
                        obj.last_modified > newest_meta.last_modified:
                    newest_meta = obj
        if newest_res is None or newest_meta is None:
            logger.warning('Found no objects for %s', pol_name)
            return {}, self.UNKNOWN_RESOURCE_TYPE
        logger.debug(
            'Found newest key for %s in %s: resource: %s, metadata: %s',
            pol_name, bucket.name, newest_res.key, newest_meta.key
        )
        # ok, ``newest`` is the newest resource.json for the policy; read it
        resource = self._extract_data_from_s3_obj(newest_res)
        _type = ''
        if get_res_type:
            metadata = self._extract_data_from_s3_obj(newest_meta)
            _type = metadata.get('policy', {}) \
                .get('resource', self.UNKNOWN_RESOURCE_TYPE)
        return resource, _type

    def _extract_data_from_s3_obj(self, obj):
        """
        Extracts a JSON payload from an S3 object.
        """
        res = obj.get()
        body = res['Body']
        if obj.key.endswith('.gz'):
            # object is gzipped; see c7n.output.FSOutput.compress()
            body = body.read()
            body = decompress(body, 15 + 32)
        return json.loads(body)


def parse_args(argv):
    p = argparse.ArgumentParser(
        description='Generate a diff of resources matched by policies in '
                    'this dryrun vs the live policies'
    )
    p.add_argument('-V', '--version', action='version', version=VERSION)
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-g', '--git-dir', dest='git_dir', action='store', type=str,
                   default=None, help='path to git directory, if not pwd')
    p.add_argument('-d', '--diff-against', dest='diff_against', action='store',
                   type=str, default='master',
                   help='git ref to diff against (default: master)')
    p.add_argument('-c', '--config', dest='config', action='store',
                   default='manheim-c7n-tools.yml',
                   help='Config file path (default: ./manheim-c7n-tools.yml)')
    p.add_argument('ACCOUNT_NAME', type=str, action='store',
                   help='Account name in config file, to run diff for')
    args = p.parse_args(argv)
    return args


def main():
    global logger
    # setup logging for direct command-line use
    FORMAT = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.WARNING, format=FORMAT)
    logger = logging.getLogger()

    # suppress boto3 internal logging below WARNING level
    boto3_log = logging.getLogger("boto3")
    boto3_log.setLevel(logging.WARNING)
    boto3_log.propagate = True

    # suppress botocore internal logging below WARNING level
    botocore_log = logging.getLogger("botocore")
    botocore_log.setLevel(logging.WARNING)
    botocore_log.propagate = True
    # end setup logging

    args = parse_args(sys.argv[1:])

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    conf = ManheimConfig.from_file(args.config, args.ACCOUNT_NAME)
    DryRunDiffer(conf).run(
        git_dir=args.git_dir,
        diff_against=args.diff_against,
    )


if __name__ == "__main__":
    main()



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import re
from copy import deepcopy
from collections import defaultdict
from datetime import datetime
from tabulate import tabulate
import argparse
import logging
import shutil

import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from manheim_c7n_tools.version import VERSION, PROJECT_URL
from manheim_c7n_tools.config import ManheimConfig
from manheim_c7n_tools.utils import git_html_url

whtspc_re = re.compile(r'\s+')

logger = logging.getLogger(__name__)


def strip_doc(func):
    """
    Given a function or method reference, return its docstring as one line (with
    all newlines removed and all whitespace collapsed).
    """
    d = func.__doc__.replace("\n", " ").strip()
    return whtspc_re.sub(' ', d)


def timestr():
    """just here to make unit testing simpler"""
    return datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S') + ' UTC'


def is_enabled(policy):
    """
    Helper function to determine if a policy is enabled.

    :param policy: policy to check
    :type policy: dict
    """
    return not(policy.get("disable", False))


class PolicyGen(object):

    def __init__(self, config):
        """
        Initialize the policy generator tool.

        :param config: manheim-c7n-tools configuration object
        :type config: ManheimConfig
        """
        self._config = config
        logger.info(
            'Initialized PolicyGen for account: %s (%s)',
            self._config.account_name, self._config.account_id
        )
        self._policy_sources = defaultdict(set)

    def run(self):
        defaults = self._load_defaults()
        if defaults is None:
            logger.error('Failed to find a `defaults.yml` file')
            raise SystemExit(1)
        acct_configs = self._load_all_policies()
        # generate the per-region configs for each region, for current account
        for rname in self._config.regions:
            self._generate_configs(
                acct_configs[self._config.account_name][rname],
                defaults,
                rname
            )
        logger.info('Writing policy descriptions to policies.rst...')
        self._write_file('policies.rst', self._policy_rst(acct_configs))
        logger.info('Writing region list to regions.rst...')
        self._write_file('regions.rst', self._regions_rst())
        self._setup_mailer_templates()

    def _load_defaults(self):
        """
        Load a defaults.yml file from either the ``policies/`` subdirectory
        or directories in the ``policy_source_paths`` configuration key.
        """
        defaults = None
        # read the global defaults
        if os.path.exists(os.path.join('policies', 'defaults.yml')):
            defaults = self._read_file_yaml(
                os.path.join('policies', 'defaults.yml')
            )

        # check policy folders for defaults
        try:
            paths = self._config.policy_source_paths
            for path in paths:
                if os.path.exists(
                    os.path.join('policies', path, 'defaults.yml')
                ):
                    defaults = self._read_file_yaml(
                        os.path.join('policies', path, 'defaults.yml')
                    )
        except AttributeError:
            logger.debug("No additional source paths for defaults")
        return defaults

    def _load_all_policies(self):
        """
        Read the policies, either the current list of ``policy_source_paths``
        directories if the config key exists, or simply the ``policies/``
        subdirectory if it doesn't.
        """
        # dict to hold account_name -> config for that account
        acct_configs = {}
        try:
            logger.info(
                "Reading from multiple source paths: %s",
                self._config.policy_source_paths
            )
            for path in self._config.policy_source_paths:
                logger.info("Reading configs from %s", path)
                configs = self._load_policy(path=path)
                # update self._policy_sources to track where each policy was
                for aname, adata in configs.items():
                    for rname, rdata in adata.items():
                        for pname in rdata.keys():
                            self._policy_sources[pname].add(path)
                acct_configs = self._merge_configs(acct_configs, configs)
                logger.info(
                    "Merging configs from %s into existing configs", path
                )
        except AttributeError:
            logger.info(
                "No source paths defined, falling back to single source path"
            )
            acct_configs = self._load_policy()
        return acct_configs

    def _merge_configs(self, target, source):
        new_config = deepcopy(target)
        for account in source:
            if account in new_config:
                for region in source[account]:
                    if region in new_config[account]:
                        for rule in source[account][region]:
                            logger.info("Rule: " + rule)
                            disable = source[account][region][rule].get(
                                "disable", False
                            )
                            if disable is True:
                                if rule in new_config[account][region]:
                                    del new_config[account][region][rule]
                            else:
                                new_config[account][region][rule] = \
                                    deepcopy(source[account][region][rule])
                    else:
                        new_config[account][region] = \
                            deepcopy(source[account][region])
            else:
                new_config[account] = deepcopy(source[account])
        return new_config

    def _load_policy(self, path=''):
        """
        Load all policies in a given path; return a nested dict of account name
        (str) to region name (str) to dict of policy names (str) to policies
        (dict).

        :param path: path to load policies from
        :type path: str
        :return: nested dict of policies
        :rtype: dict
        """
        acct_configs = {}
        # read the shared configs from all_accounts/ ; returns a dict of
        # region name to [dict of policy name to policy], for each region
        all_accts = self._read_policy_directory(
            os.path.join(path, 'all_accounts')
        )
        # loop over all accounts in the config file
        for acctname in self._config.list_accounts(self._config.config_path):
            # start with the all_accts dict, for common config
            conf = deepcopy(all_accts)
            # read the account's config
            acct_conf = self._read_policy_directory(
                os.path.join(path, acctname)
            )
            # for each region, layer per-account over all_accounts
            for rname in self._config.regions:
                conf[rname].update(acct_conf[rname])
            acct_configs[acctname] = deepcopy(conf)
        return acct_configs

    def _read_policy_directory(self, policy_dir):
        """
        Read all policies from a ``policies/`` subdirectory (``all_accounts/``
        or an account name). Return a dict of region names to dict of policies
        (name to policy) for that region.

        :param policy_dir: ``policies/`` subdirectory name to read policies from
        :type policy_dir: str
        :return: dict of region name to policies dict (name to policy)
        :rtype: dict
        """
        common = self._read_policies(os.path.join(policy_dir, 'common'))
        region_policies = {}
        for rname in self._config.regions:
            policies = deepcopy(common)
            policies.update(
                self._read_policies(os.path.join(policy_dir, rname))
            )
            region_policies[rname] = policies
        return region_policies

    def _generate_configs(self, policies, defaults, region_name):
        """
        Given policies read from disk, apply defaults, generate cleanup
        policies, sanity/safety check policies. Then write the custodian configs
        to disk and return the resulting policies dict.

        :param policies: the policies read from disk (return value of
          :py:meth:`~._read_policies`)
        :type policies: dict
        :param defaults: the defaults to apply to the policies
        :type defaults: dict
        :param region_name: the name of the region these configs are for
        :type region_name: str
        :return: dictionary of final policies
        :rtype: dict
        """
        result = {'policies': []}
        for k in sorted(policies.keys()):
            result['policies'].append(
                self._apply_defaults(defaults, policies[k])
            )
        if self._config.cleanup_notify:
            logger.info('Generating c7n cleanup policies...')
            # add c7n lambda/CW Event cleanup policies
            for pol in self._generate_cleanup_policies(
                deepcopy(result['policies'])
            ):
                result['policies'].append(self._apply_defaults(defaults, pol))
        logger.info('Checking policies for sanity and safety...')
        self._check_policies(result['policies'])
        self._write_custodian_configs(result, region_name)
        return result

    def _write_custodian_configs(self, result, region_name):
        """
        Write the per-region ``custodian_REGION.yml`` config file to disk. This
        also handles ``%%`` macro and environment variable substitution.

        :param result: final custodian configuration
        :type result: dict
        :param region_name: the name of the region the configs are for
        :type region_name: str
        """
        enabled_policies = list(filter(is_enabled, result['policies']))
        config_str = yaml.dump({"policies": enabled_policies})
        fname = 'custodian_%s.yml' % region_name
        logger.info('Writing %s policies to %s...' % (region_name, fname))
        conf = config_str
        replacements = [
            ['%%BUCKET_NAME%%', self._config.output_s3_bucket_name],
            ['%%LOG_GROUP%%', self._config.custodian_log_group],
            ['%%DLQ_ARN%%', self._config.dead_letter_queue_arn],
            ['%%ROLE_ARN%%', self._config.role_arn],
            ['%%MAILER_QUEUE_URL%%', self._config.mailer_config['queue_url']],
            ['%%ACCOUNT_NAME%%', self._config.account_name],
            ['%%ACCOUNT_ID%%', str(self._config.account_id)],
            ['%%AWS_REGION%%', region_name]
        ]
        for k, v in os.environ.items():
            if k.startswith('POLICYGEN_ENV_'):
                replacements.append(['%%' + k + '%%', v])
        for macro, val in replacements:
            conf = conf.replace(macro, val)
        self._write_file(fname, conf)

    def _check_policies(self, policies):
        """
        Check all of our policies to ensure that they conform with some rules
        and best practices around safety and sanity.

        Each policy in ``policies`` is passed through each of the
        ``self._check_policy_*`` functions (which return a boolean pass/fail).
        At the end, all failures are collected. If there are any, SystemExit(1)
        is raised.

        :param policies: list of policy dictionaries
        :type policies: list
        :raises: SystemExit(1) if any policies failed checks
        """
        policy_checks = []
        for x in dir(self):
            if x.startswith('_check_policy_') and callable(getattr(self, x)):
                policy_checks.append(getattr(self, x))
        failures = defaultdict(list)
        for pol in policies:
            for chk in policy_checks:
                if not chk(pol):
                    failures[pol['name']].append(strip_doc(chk))
        if len(failures) > 0:
            logger.error('ERROR: Some policies failed sanity/safety checks:')
            for pol_name in sorted(failures.keys()):
                logger.error(pol_name)
                for chk_str in failures[pol_name]:
                    logger.error("\t" + chk_str)
            raise SystemExit(1)
        logger.info('OK: All policies passed sanity/safety checks.')

    def _check_policy_function_prefix(self, policy):
        """
        Fail if function-prefix doesn't match between manheim-c7n-tools config
        and the policy.
        """
        fp = policy.get('mode', {}).get('function-prefix', 'custodian-')
        if fp != self._config.function_prefix:
            return False
        return True

    def _check_policy_marked_for_op_first(self, policy):
        """
        Policy includes a marked-for-op filter, but it is not the first filter.
        """
        if 'filters' not in policy:
            return True
        if "'type': 'marked-for-op'" not in str(policy['filters']):
            return True
        try:
            if policy['filters'][0].get('type', '') == 'marked-for-op':
                return True
        except AttributeError:
            # first filter isn't even a dict; that's a failure
            pass
        # fail - first filter isn't marked-for-op
        return False

    def _check_policy_mark_but_no_tag_filter(self, policy):
        """
        Policy performs a mark action, but does not filter out resources already
        marked with that tag.
        """
        if 'filters' not in policy:
            return True
        if 'actions' not in policy:
            return True
        mark_tags = []
        for a in policy['actions']:
            if not isinstance(a, type({})):
                # not a dict, can't be a mark action
                continue
            if a.get('type', '') != 'mark-for-op':
                continue
            mark_tags.append(a['tag'])
        for t in mark_tags:
            tag_filter = {'tag:%s' % t: 'absent'}
            if tag_filter not in policy['filters']:
                return False
        return True

    def _check_policy_mark_for_op_bad_message(self, policy):
        """
        mark-for-op action has message that does not end with
        ": {op}@{action_date}" (won't be parsed by c7n and will be ignored)
        """
        success = True
        if 'actions' not in policy:
            return True
        for a in policy['actions']:
            if not isinstance(a, type({})):
                # not a dict, can't be a mark action
                continue
            if a.get('type', '') != 'mark-for-op':
                continue
            if 'message' not in a:
                continue
            if not a['message'].endswith(': {op}@{action_date}'):
                success = False
        return success

    def _generate_cleanup_policies(self, policies):
        """
        When c7n is run, it provisions all policies as lambda functions. But if
        policies are removed, it doesn't know how to clean them up. See
        https://github.com/capitalone/cloud-custodian/issues/48

        As a workaround for this, we tag all Lambda funcs created by c7n
        with Project: cloud-custodian and a Component tag of the policy name.

        This method generates policies that look for cloud-custodian Lambda
        functions and CloudWatch Events that aren't in the current list of
        policies, and therefore probably need cleanup, and notifies us.

        :param policies: list of policy dictionaries
        :type policies: list
        :return: list of c7n cleanup policies to add
        :rtype: list
        """
        # base policies that just need filters added
        lcleanup = {
            'name': 'c7n-cleanup-lambda',
            'comment': 'Find and alert on orphaned c7n Lambda functions',
            'resource': 'lambda',
            'actions': [{
                'type': 'notify',
                'violation_desc': 'The following cloud-custodian Lambda '
                                  'functions appear to be orphaned',
                'action_desc': 'and should probably be deleted',
                'subject': '[cloud-custodian {{ account }}] Orphaned '
                           'cloud-custodian Lambda funcs in {{ region }}',
                'to': self._config.cleanup_notify,
            }],
            'filters': [
                {'tag:Project': 'cloud-custodian'},
                {'tag:Component': 'present'},
                # exclude itself...
                {
                    'type': 'value',
                    'key': 'tag:Component',
                    'op': 'ne',
                    'value': 'c7n-cleanup-lambda'
                },
                {
                    'type': 'value',
                    'key': 'tag:Component',
                    'op': 'ne',
                    'value': 'c7n-cleanup-cwe'
                }
            ]
        }
        cwecleanup = {
            'name': 'c7n-cleanup-cwe',
            'comment': 'Find and alert on orphaned c7n CloudWatch Events',
            'resource': 'event-rule',
            'actions': [{
                'type': 'notify',
                'violation_desc': 'The following cloud-custodian CloudWatch '
                                  'Event rules appear to be orphaned',
                'action_desc': 'and should probably be deleted',
                'subject': '[cloud-custodian {{ account }}] Orphaned '
                           'cloud-custodian CW Event rules in {{ region }}',
                'to': self._config.cleanup_notify,
            }],
            'filters': [
                {
                    'type': 'value',
                    'key': 'Name',
                    'op': 'glob',
                    'value': 'custodian-*'
                },
                # exclude itself...
                {
                    'type': 'value',
                    'key': 'Name',
                    'op': 'ne',
                    'value': 'custodian-c7n-cleanup-lambda'
                },
                {
                    'type': 'value',
                    'key': 'Name',
                    'op': 'ne',
                    'value': 'custodian-c7n-cleanup-cwe'
                }
            ]
        }
        # add the filters
        for p in policies:
            name = p['name']
            if is_enabled(p):
                cwecleanup['filters'].append({
                    'type': 'value',
                    'key': 'Name',
                    'op': 'ne',
                    'value': 'custodian-%s' % name
                })
                lcleanup['filters'].append({
                    'type': 'value',
                    'key': 'tag:Component',
                    'op': 'ne',
                    'value': name
                })
        return [lcleanup, cwecleanup]

    def _write_file(self, path, content):
        """write a file - helper to make unit tests simpler"""
        with open(path, 'w') as fh:
            fh.write(content)

    def _apply_defaults(self, defaults, policy):
        d = deepcopy(defaults)
        conf = self._merge_conf(d, policy, policy['name'], [])
        # set Lambda func 'Component' tag to the policy name
        if conf['mode']['type'] == 'periodic' and 'tags' not in conf['mode']:
            conf['mode']['tags'] = {}
        if 'tags' in conf['mode']:
            conf['mode']['tags'].update(
                defaults.get('mode', {}).get('tags', {})
            )
            conf['mode']['tags']['Component'] = policy['name']
        if 'actions' not in conf:
            conf['actions'] = []
        return self._add_always_notify(conf)

    def _add_always_notify(self, conf):
        """
        Given a policy configuration like the one returned by
        :py:meth:`~._apply_defaults`, return the input unchanged if the
        ``always_notify`` configuration value is empty or not present, or else
        ensure that the policy contains at least one ``type: notify`` action
        with the specified transport and ``to``.

        :param conf: configuration as returned by :py:meth:`~._apply_defaults`
        :type conf: dict
        :return: conf with always_notify action if configured
        :rtype: dict
        """
        try:
            desired = self._config.always_notify
            desired['type'] = 'notify'
        except AttributeError:
            return conf
        added = False
        for action in conf['actions']:
            if not isinstance(action, type({})):
                continue
            if action.get('type', None) != 'notify':
                continue
            if action.get('transport', {}) != desired['transport']:
                continue
            # ok, we've got an action with the desired transport
            if 'to' not in action:
                action['to'] = []
            for to_addr in desired['to']:
                if to_addr not in action['to']:
                    action['to'].append(to_addr)
            added = True
            break
        if not added:
            conf['actions'].append(desired)
        return conf

    def _merge_conf(self, base, update, policy_name, path):
        """merge update into base"""
        for k, v in update.items():
            kpath = path + [k]
            if (
                kpath == ['mode'] and v.get('type', 'periodic') != 'periodic'
            ):
                # short-circuit to not alter the 'mode' top-level key on
                # policies if it isn't "type: periodic"
                base[k] = update[k]
                continue
            if k not in base:
                base[k] = v
                continue
            if isinstance(v, type([])):
                # List / array
                base[k] = self._array_merge(base[k], v, policy_name, kpath)
            elif isinstance(v, type({})):
                # nested dictionary
                base[k] = self._merge_conf(base[k], v, policy_name, kpath)
            else:
                # not a dict or list; probably string or int, etc.
                base[k] = v
        # remove actions if only specified in base (defaults)
        if path == [] and 'actions' in base and 'actions' not in update:
            del base['actions']
        return base

    def _array_merge(self, base, update, policy_name, path):
        """this starts with update, and adds things from base"""
        if not isinstance(base, type([])):
            logger.error(
                'ERROR: policy has an array but defaults does not; cannot merge'
            )
            raise RuntimeError(
                'Policy %s: Cannot array merge non-array from defaults (%s)' % (
                    policy_name, base
                )
            )
        # find the defaults, by type
        def_dicts = {}
        for v in base:
            # coverage doesn't report correctly on this branch
            if not isinstance(v, type({})):
                if v not in update:  # nocoverage
                    update.append(v)  # nocoverage
                continue  # nocoverage
            t = v.get('type', None)
            if t is None:
                raise RuntimeError('Do not know how to handle a defaults '
                                   'dict without a "type" key.')
            if t in def_dicts:
                raise RuntimeError('Defaults cannot specify multiple dicts '
                                   'with the same "type" in the same array!')
            def_dicts[t] = v
        # do the updates
        for i in update:
            if not isinstance(i, type({})):
                continue
            # else it's a dict, update from defaults if present
            t = i.get('type', None)
            if t is None:
                continue
            if t not in def_dicts:
                # no defaults for this
                continue
            for k, v in def_dicts[t].items():
                if k not in i:
                    i[k] = v
            del def_dicts[t]
        # add any defaults that didn't already exist
        for k, v in def_dicts.items():
            if path == ['actions'] and k == 'notify':
                # Don't add notify actions to policies that don't have them
                continue
            update.append(v)
        return update

    def _policy_rst(self, region_policies):
        """
        Build the policies rST source for the documentation.

        :param region_policies: dict of region names to per-region dict of
          policy name to policy content, for that region.
        :type region_policies: dict
        :return: built rST markup for policies docs
        :rtype: str
        """
        buildinfo = 'by `%s %s <%s>`_' % (
            os.environ.get('JOB_NAME', ''),
            os.environ.get('BUILD_NUMBER', ''),
            os.environ.get('BUILD_URL', '')
        )
        commit = os.environ.get('GIT_COMMIT', 'unknown')
        gitlink = '%scommit/%s' % (git_html_url(), commit)
        if buildinfo == 'by `  <>`_':
            buildinfo = 'locally'
        s = "this page built %s from `%s <%s>`_ at %s\n\n" % (
            buildinfo, commit, gitlink, timestr()
        )
        try:
            assert len(self._config.policy_source_paths) > 0
            headers = [
                'Policy Name', 'Account(s) / Region(s)', 'Source Path(s)',
                'Description/Comment', 'Enabled'
            ]
            have_source_paths = True
        except Exception:
            headers = [
                'Policy Name', 'Account(s) / Region(s)', 'Description/Comment',
                'Enabled'
            ]
            have_source_paths = False
        s += tabulate(
            sorted(self._policy_rst_data(
                region_policies, have_paths=have_source_paths
            )),
            headers=headers, tablefmt='grid'
        )
        return s

    def _policy_rst_data(self, account_policies, have_paths=False):
        """
        Build the policy rST table data.

        :param account_policies: dict of Account names to dict of [region names
          to per-region dict of policy name to policy content].
        :type account_policies: dict
        :return: list of [name, regions, comment] lists for each policy
        :rtype: ``list``
        """
        all_regions = sorted(self._config.regions)
        acct_names = sorted(account_policies.keys())
        names_to_accts_regions = {
            x: defaultdict(list) for x in account_policies.keys()
        }
        metadata = {}
        for acctname in sorted(account_policies.keys()):
            region_policies = account_policies[acctname]
            for rname in sorted(region_policies.keys()):
                policies = region_policies[rname]
                for pname in sorted(policies.keys()):
                    names_to_accts_regions[acctname][pname].append(rname)
                    metadata[pname] = {
                        "description": self._policy_comment(policies[pname]),
                        "enabled": is_enabled(policies[pname])
                    }
        result = []
        for pname in sorted(metadata.keys()):
            accts = []
            for acctname in acct_names:
                regions = sorted(names_to_accts_regions[acctname][pname])
                if regions == all_regions:
                    accts.append(acctname)
                elif len(regions) > 0:
                    accts.append('%s (%s)' % (
                        acctname, ' '.join(regions)
                    ))
            if accts == acct_names:
                apart = ''
            else:
                apart = ' '.join(accts)
            if have_paths:
                result.append([
                    pname,
                    apart,
                    ' '.join(sorted(self._policy_sources.get(pname, []))),
                    metadata[pname]["description"],
                    metadata[pname]["enabled"]
                ])
            else:
                result.append([
                    pname,
                    apart,
                    metadata[pname]["description"],
                    metadata[pname]["enabled"]
                ])
        return result

    def _regions_rst(self):
        res = ''
        for acctname in self._config.list_accounts(self._config.config_path):
            aconf = self._config.from_file(self._config.config_path, acctname)
            res += '  * %s (%s)\n\n' % (acctname, aconf.account_id)
            res += "\n".join(['    * %s' % r for r in aconf.regions]) + "\n\n"
        return res

    def _policy_comment(self, policy):
        for k in ['comment', 'comments', 'description']:
            if k in policy:
                return policy[k].strip()
        return 'unknown'

    def _read_policies(self, subdir):
        """
        Read policy files from a subdirectory of the policies directory, and
        return the resulting dict of policy names to policy contents.

        :param subdir: directory path under ``policies/`` to read
        :type subdir: str
        :return: dict of policy names to policies
        :rtype: dict
        """
        res = {}
        try:
            for f in os.listdir(os.path.join('policies', subdir)):
                if not f.endswith('.yml'):
                    continue
                name = f.split('.')[0]
                y = self._read_file_yaml(os.path.join('policies', subdir, f))
                res[name] = y
                if name != 'defaults' and y.get('name', '') != name:
                    raise RuntimeError(
                        'ERROR: Policy file %s contains policy with name '
                        '"%s".' % (f, y.get('name', ''))
                    )
        except OSError:
            return {}
        logger.info(
            'Loaded %d policies from %s: %s', len(res), subdir, res.keys()
        )
        return res

    def _read_file_yaml(self, path):
        """unit test helper - return YAML from file contents"""
        with open(path, 'r') as fh:
            contents = fh.read()
        try:
            return yaml.load(contents, Loader=SafeLoader)
        except Exception:
            sys.stderr.write("Exception loading YAML: %s\n" % path)
            raise

    def _setup_mailer_templates(self):
        """
        Call :py:meth:`~._mailer_template_paths`. If it returns an empty dict,
        do nothing. Otherwise, create ``./mailer-templates`` if it does not
        already exist. For each template filename that does not already exist
        in that directory, copy it from the source path specified by
        :py:meth:`~._mailer_template_paths`.
        """
        paths = self._mailer_template_paths()
        if not paths:
            logger.info(
                'No mailer-templates directories found in policy_source_paths; '
                'not setting up mailer templates.'
            )
            return
        if not os.path.exists('mailer-templates'):
            logger.info('Creating directory: mailer-templates')
            os.mkdir('mailer-templates')
        for fname, srcpath in paths.items():
            destpath = os.path.join('mailer-templates', fname)
            if os.path.exists(destpath):
                logger.info(
                    '%s already exists in pwd; not overwriting', destpath
                )
            else:
                logger.info(
                    'Setting up %s from source at %s', destpath, srcpath
                )
                shutil.copyfile(srcpath, destpath)
        logger.info('Done setting up mailer-templates')

    def _mailer_template_paths(self):
        """
        Find all files in the ``mailer-templates`` subdirectory of each
        ``policy_source_paths`` directory, if present. Return a dictionary of
        file name to file path. If a file with the same name is found in
        multiple directories, the last one in ``policy_source_paths`` order
        wins.

        :return: Mailer template names to their source paths
        :rtype: dict
        """
        templates = {}
        try:
            logger.debug(
                "Finding mailer-templates in: %s",
                self._config.policy_source_paths
            )
            for path in self._config.policy_source_paths:
                mailerdir = os.path.join('policies', path, 'mailer-templates')
                if not os.path.exists(mailerdir):
                    logger.debug('%s does not exist; skipping', mailerdir)
                    continue
                logger.info('Finding mailer templates in %s', mailerdir)
                for f in os.listdir(mailerdir):
                    fpath = os.path.join(mailerdir, f)
                    if not os.path.isfile(fpath):
                        continue
                    logger.debug(
                        'Using mailer template %s from %s', f, fpath
                    )
                    templates[f] = fpath
        except AttributeError:
            logger.debug(
                "No policy_source_paths defined; not setting up "
                "mailer-templates."
            )
        return templates


def main():
    # setup logging for direct command-line use
    global logger
    FORMAT = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.INFO, format=FORMAT)
    logger = logging.getLogger()

    # suppress boto3 internal logging below WARNING level
    boto3_log = logging.getLogger("boto3")
    boto3_log.setLevel(logging.WARNING)
    boto3_log.propagate = True

    # suppress botocore internal logging below WARNING level
    botocore_log = logging.getLogger("botocore")
    botocore_log.setLevel(logging.WARNING)
    botocore_log.propagate = True

    p = argparse.ArgumentParser(
        description='Tool to generate custodian config files '
                    'from a c7n configuration directory/repo.',
        epilog='This tool is part of manheim_c7n_tools v%s.\n'
               'For documentation, see: %s' % (VERSION, PROJECT_URL),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    p.add_argument('-V', '--version', action='version', version=VERSION)
    p.add_argument('-c', '--config', dest='config', action='store',
                   default='manheim-c7n-tools.yml',
                   help='Config file path (default: ./manheim-c7n-tools.yml)')
    p.add_argument('ACCT_NAME', action='store', type=str,
                   help='account_name value from config file, for '
                        'current account')

    args = p.parse_args(sys.argv[1:])
    conf = ManheimConfig.from_file(args.config, args.ACCT_NAME)
    PolicyGen(conf).run()


if __name__ == "__main__":
    main()



"""
This file was copied verbatim from the cloud-custodian source,
``tools/ops/mugc.py``, as of the 0.9.1.0 tag. They're not included in the
custodian Python package, only in the git repo, so we need to vendor them in.
"""
# Copyright 2016-2018 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import itertools
import json
import os
import re
import logging
import sys

from c7n.credentials import SessionFactory
from c7n.config import Config
from c7n.policy import load as policy_load, PolicyCollection
from c7n import mu

# TODO: mugc has alot of aws assumptions

from c7n.resources.aws import AWS
import boto3
from botocore.exceptions import ClientError


log = logging.getLogger('mugc')


def load_policies(options, config):
    policies = PolicyCollection([], config)
    for f in options.config_files:
        policies += policy_load(config, f).filter(options.policy_filter)
    return policies


def region_gc(options, region, policy_config, policies):

    session_factory = SessionFactory(
        region=region,
        assume_role=policy_config.assume_role,
        profile=policy_config.profile,
        external_id=policy_config.external_id)

    manager = mu.LambdaManager(session_factory)
    funcs = list(manager.list_functions(options.prefix))
    client = session_factory().client('lambda')

    remove = []
    current_policies = [p.name for p in policies]
    pattern = re.compile(options.policy_regex)
    for f in funcs:
        if not pattern.match(f['FunctionName']):
            continue
        match = False
        for pn in current_policies:
            if f['FunctionName'].endswith(pn):
                match = True
        if options.present:
            if match:
                remove.append(f)
        elif not match:
            remove.append(f)

    for n in remove:
        events = []
        try:
            result = client.get_policy(FunctionName=n['FunctionName'])
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                log.warning(
                    "Region:%s Lambda Function or Access Policy Statement missing: %s",
                    region, n['FunctionName'])
            else:
                log.warning(
                    "Region:%s Unexpected error: %s for function %s",
                    region, e, n['FunctionName'])

            # Continue on with next function instead of raising an exception
            continue

        if 'Policy' not in result:
            pass
        else:
            p = json.loads(result['Policy'])
            for s in p['Statement']:
                principal = s.get('Principal')
                if not isinstance(principal, dict):
                    log.info("Skipping function %s" % n['FunctionName'])
                    continue
                if principal == {'Service': 'events.amazonaws.com'}:
                    events.append(
                        mu.CloudWatchEventSource({}, session_factory))
                elif principal == {'Service': 'config.amazonaws.com'}:
                    events.append(
                        mu.ConfigRule({}, session_factory))

        f = mu.LambdaFunction({
            'name': n['FunctionName'],
            'role': n['Role'],
            'handler': n['Handler'],
            'timeout': n['Timeout'],
            'memory_size': n['MemorySize'],
            'description': n['Description'],
            'runtime': n['Runtime'],
            'events': events}, None)

        log.info("Region:%s Removing %s", region, n['FunctionName'])
        if options.dryrun:
            log.info("Dryrun skipping removal")
            continue
        manager.remove(f)
        log.info("Region:%s Removed %s", region, n['FunctionName'])


def resources_gc_prefix(options, policy_config, policy_collection):
    """Garbage collect old custodian policies based on prefix.

    We attempt to introspect to find the event sources for a policy
    but without the old configuration this is implicit.
    """

    # Classify policies by region
    policy_regions = {}
    for p in policy_collection:
        if p.execution_mode == 'poll':
            continue
        policy_regions.setdefault(p.options.region, []).append(p)

    regions = get_gc_regions(options.regions)
    for r in regions:
        region_gc(options, r, policy_config, policy_regions.get(r, []))


def get_gc_regions(regions):
    if 'all' in regions:
        session = boto3.Session(
            region_name='us-east-1',
            aws_access_key_id='never',
            aws_secret_access_key='found')
        return session.get_available_regions('s3')
    return regions


def setup_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("configs", nargs='*', help="Policy configuration file(s)")
    parser.add_argument(
        '-c', '--config', dest="config_files", nargs="*", action='append',
        help="Policy configuration files(s)", default=[])
    parser.add_argument(
        "--present", action="store_true", default=False,
        help='Target policies present in config files for removal instead of skipping them.')
    parser.add_argument(
        '-r', '--region', action='append', dest='regions', metavar='REGION',
        help="AWS Region to target. Can be used multiple times, also supports `all`")
    parser.add_argument('--dryrun', action="store_true", default=False)
    parser.add_argument(
        "--profile", default=os.environ.get('AWS_PROFILE'),
        help="AWS Account Config File Profile to utilize")
    parser.add_argument(
        "--prefix", default="custodian-",
        help="The Lambda name prefix to use for clean-up")
    parser.add_argument(
        "--policy-regex",
        help="The policy must match the regex")
    parser.add_argument("-p", "--policies", default=None, dest='policy_filter',
                        help="Only use named/matched policies")
    parser.add_argument(
        "--assume", default=None, dest="assume_role",
        help="Role to assume")
    parser.add_argument(
        "-v", dest="verbose", action="store_true", default=False,
        help='toggle verbose logging')
    return parser


def main():
    parser = setup_parser()
    options = parser.parse_args()

    log_level = logging.INFO
    if options.verbose:
        log_level = logging.DEBUG
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s: %(name)s:%(levelname)s %(message)s")
    logging.getLogger('botocore').setLevel(logging.ERROR)
    logging.getLogger('urllib3').setLevel(logging.ERROR)
    logging.getLogger('c7n.cache').setLevel(logging.WARNING)

    if not options.policy_regex:
        options.policy_regex = f"^{options.prefix}.*"

    if not options.regions:
        options.regions = [os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')]

    files = []
    files.extend(itertools.chain(*options.config_files))
    files.extend(options.configs)
    options.config_files = files

    if not files:
        parser.print_help()
        sys.exit(1)

    policy_config = Config.empty(
        regions=options.regions,
        profile=options.profile,
        assume_role=options.assume_role)

    # use cloud provider to initialize policies to get region expansion
    policies = AWS().initialize_policies(
        PolicyCollection([
            p for p in load_policies(
                options, policy_config)
            if p.provider_name == 'aws'],
            policy_config),
        policy_config)

    resources_gc_prefix(options, policy_config, policies)


if __name__ == '__main__':
    main()



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import patch, call, Mock, mock_open
import pytest
import yaml

from manheim_c7n_tools.config import ManheimConfig, MANHEIM_CONFIG_SCHEMA

pbm = 'manheim_c7n_tools.config'


class TestManheimConfig(object):

    def test_init(self):
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch(
                '%s.jsonschema.validate' % pbm, autospec=True
            ) as mock_validate:
                cls = ManheimConfig(
                    foo='bar', baz=2, regions=['us-east-1'],
                    config_path='manheim-c7n-tools.yml', account_id='1234',
                    cleanup_notify=['foo@bar.com']
                )
        assert cls._config == {
            'foo': 'bar', 'baz': 2, 'regions': ['us-east-1'],
            'account_id': '1234', 'cleanup_notify': ['foo@bar.com'],
            'function_prefix': 'custodian-'
        }
        assert cls.config_path == 'manheim-c7n-tools.yml'
        assert mock_logger.mock_calls == [
            call.debug('Validating configuration...')
        ]
        assert mock_validate.mock_calls == [
            call(
                {
                    'foo': 'bar', 'baz': 2, 'regions': ['us-east-1'],
                    'account_id': '1234', 'cleanup_notify': ['foo@bar.com']
                },
                MANHEIM_CONFIG_SCHEMA
            )
        ]

    def test_init_not_us_east_1(self):
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch(
                '%s.jsonschema.validate' % pbm, autospec=True
            ) as mock_validate:
                cls = ManheimConfig(
                    foo='bar', baz=2, regions=['us-east-2'],
                    config_path='manheim-c7n-tools.yml', account_id='1234',
                    cleanup_notify=['foo@bar.com'], function_prefix='foo-'
                )
        assert cls._config == {
            'foo': 'bar', 'baz': 2, 'regions': ['us-east-2'],
            'account_id': '1234', 'cleanup_notify': ['foo@bar.com'],
            'function_prefix': 'foo-'
        }
        assert cls.config_path == 'manheim-c7n-tools.yml'
        assert mock_logger.mock_calls == [
            call.debug('Validating configuration...')
        ]
        assert mock_validate.mock_calls == [
            call(
                {
                    'foo': 'bar', 'baz': 2, 'regions': ['us-east-2'],
                    'account_id': '1234', 'cleanup_notify': ['foo@bar.com'],
                    'function_prefix': 'foo-'
                },
                MANHEIM_CONFIG_SCHEMA
            )
        ]

    def test_getattr(self):
        with patch('%s.logger' % pbm, autospec=True):
            with patch('%s.jsonschema.validate' % pbm, autospec=True):
                cls = ManheimConfig(
                    foo='bar', baz=2, regions=['us-east-1'], config_path='foo',
                    account_id='012345'
                )
        assert cls.foo == 'bar'
        assert cls.baz == 2
        assert cls.config_path == 'foo'
        assert cls.account_id == '012345'
        with pytest.raises(AttributeError):
            cls.missingAttr

    def test_from_file(self):
        m_conf = Mock()
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch(
                '%s.open' % pbm, mock_open(read_data='foo'), create=True
            ) as m_open:
                with patch('%s.yaml.load' % pbm, autospec=True) as mock_load:
                    with patch(
                        '%s.ManheimConfig' % pbm, autospec=True
                    ) as mock_conf:
                        mock_conf.return_value = m_conf
                        mock_load.return_value = [
                            {
                                'account_name': 'a1',
                                'foo': 'bar',
                                'baz': 2,
                                'regions': ['us-east-1']
                            },
                            {
                                'account_name': 'a2',
                                'foo': 'bar1',
                                'baz': 4,
                                'regions': ['us-east-2']
                            }
                        ]
                        res = ManheimConfig.from_file('/tmp/conf.yml', 'a2')
        assert res == m_conf
        assert mock_logger.mock_calls == [
            call.info('Loading config from: %s', '/tmp/conf.yml')
        ]
        assert m_open.mock_calls == [
            call('/tmp/conf.yml', 'r'),
            call().__enter__(),
            call().read(),
            call().__exit__(None, None, None)
        ]
        assert mock_load.mock_calls == [
            call('foo', Loader=yaml.SafeLoader)
        ]
        assert mock_conf.mock_calls == [
            call(
                account_name='a2', foo='bar1', baz=4, regions=['us-east-2'],
                config_path='/tmp/conf.yml'
            )
        ]

    def test_from_file_name_missing(self):
        m_conf = Mock()
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch(
                '%s.open' % pbm, mock_open(read_data='foo'), create=True
            ) as m_open:
                with patch('%s.yaml.load' % pbm, autospec=True) as mock_load:
                    with patch(
                        '%s.ManheimConfig' % pbm, autospec=True
                    ) as mock_conf:
                        mock_conf.return_value = m_conf
                        mock_load.return_value = [
                            {
                                'account_name': 'a1',
                                'foo': 'bar',
                                'baz': 2,
                                'regions': ['us-east-1']
                            },
                            {
                                'account_name': 'a2',
                                'foo': 'bar1',
                                'baz': 4,
                                'regions': ['us-east-2']
                            }
                        ]
                        with pytest.raises(RuntimeError) as exc:
                            ManheimConfig.from_file('/tmp/conf.yml', 'BAD')
        assert str(exc.value) == 'ERROR: No account with name "BAD"' \
                                 ' in /tmp/conf.yml'
        assert mock_logger.mock_calls == [
            call.info('Loading config from: %s', '/tmp/conf.yml')
        ]
        assert m_open.mock_calls == [
            call('/tmp/conf.yml', 'r'),
            call().__enter__(),
            call().read(),
            call().__exit__(None, None, None)
        ]
        assert mock_load.mock_calls == [
            call('foo', Loader=yaml.SafeLoader)
        ]
        assert mock_conf.mock_calls == []

    def test_list_accounts(self):
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch(
                '%s.open' % pbm, mock_open(read_data='foo'), create=True
            ) as m_open:
                with patch('%s.yaml.load' % pbm, autospec=True) as mock_load:
                    with patch(
                        '%s.ManheimConfig' % pbm, autospec=True
                    ) as mock_conf:
                        mock_load.return_value = [
                            {
                                'account_name': 'a1',
                                'account_id': '1111',
                                'foo': 'bar',
                                'baz': 2,
                                'regions': ['us-east-1']
                            },
                            {
                                'account_name': 'a2',
                                'account_id': 2222,
                                'foo': 'bar1',
                                'baz': 4,
                                'regions': ['us-east-2']
                            }
                        ]
                        res = ManheimConfig.list_accounts('/tmp/conf.yml')
        assert res == {'a1': '1111', 'a2': '2222'}
        assert mock_logger.mock_calls == [
            call.info('Loading config from: %s', '/tmp/conf.yml')
        ]
        assert m_open.mock_calls == [
            call('/tmp/conf.yml', 'r'),
            call().__enter__(),
            call().read(),
            call().__exit__(None, None, None)
        ]
        assert mock_load.mock_calls == [
            call('foo', Loader=yaml.SafeLoader)
        ]
        assert mock_conf.mock_calls == []

    def test_config_for_region(self):
        original = {
            'foo': 'bar%%AWS_REGION%%baz',
            'bar': [
                'baz',
                'AWS_REGION',
                '%%AWS_REGION%%',
                'xx%%AWS_REGION%%xx',
                'blam',
                'xx%%POLICYGEN_ENV_foo%%xx'
            ],
            'baz': {
                'blam': {
                    'blarg%%AWS_REGION%%xx': 'xxx%%AWS_REGION%%xxx'
                }
            },
            'regions': ['us-east-1', 'us-east-2'],
            'account_id': '012345',
            'config_path': '/tmp/baz.yml'
        }
        expected = {
            'foo': 'barus-east-2baz',
            'bar': [
                'baz',
                'AWS_REGION',
                'us-east-2',
                'xxus-east-2xx',
                'blam',
                'xxbarVARxx'
            ],
            'baz': {
                'blam': {
                    'blargus-east-2xx': 'xxxus-east-2xxx'
                }
            },
            'account_id': '012345',
            'regions': ['us-east-1', 'us-east-2'],
            'cleanup_notify': [],
            'function_prefix': 'custodian-'
        }
        with patch('%s.jsonschema.validate' % pbm, autospec=True):
            with patch.dict(
                'os.environ',
                {'foo': 'bar', 'POLICYGEN_ENV_foo': 'barVAR'},
                clear=True
            ):
                conf = ManheimConfig(**original)
                result = conf.config_for_region('us-east-2')
        assert result._config == expected
        assert result.config_path == '/tmp/baz.yml'



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Script to scan CloudWatch metrics, logs, and Dead Letter SQS queue for all
cloud-custodian lambda functions, and print info and logs and exit non-zero if
any failed or errored.
"""

import sys
import argparse
import logging
import boto3
from botocore.config import Config
import re
from time import time, sleep
from datetime import datetime, timedelta, tzinfo
from operator import itemgetter

from manheim_c7n_tools.utils import (
    set_log_info, set_log_debug, red, green, assume_role
)
from manheim_c7n_tools.version import VERSION, PROJECT_URL
from manheim_c7n_tools.config import ManheimConfig

FORMAT = "[%(asctime)s %(levelname)s] %(message)s"
logging.basicConfig(level=logging.WARNING, format=FORMAT)
logger = logging.getLogger()

# suppress boto3 internal logging below WARNING level
boto3_log = logging.getLogger("boto3")
boto3_log.setLevel(logging.WARNING)
boto3_log.propagate = True

# suppress botocore internal logging below WARNING level
botocore_log = logging.getLogger("botocore")
botocore_log.setLevel(logging.WARNING)
botocore_log.propagate = True

# Override max attempts for botocore retry configuration, to cope with
# throttling. This  constant is used in two different places below...
BOTOCORE_MAX_ATTEMPTS = 10


class UTC(tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


class LambdaHealthChecker(object):
    """Class for checking Lambda func health via CloudWatch"""

    # If a log message meets this exact regex, skip over it
    NO_OWNER_EMAIL_LOOKUP_WARNING = re.compile(
        r'.*(ERROR|WARNING).*unable to lookup owner email.*'
        'Please configure LDAP or org_domain')

    req_id_re = re.compile(
        r'^(START|END|REPORT|\S+\s\S+)\s'
        r'([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}).*'
    )

    def __init__(self, func_name, region_name, logs=None, cw=None):
        """
        Initialize LambdaHealthChecker

        :param func_name: Lambda function name
        :type func_name: str
        :param region_name: name of the region to run against
        :type region_name: str
        :param logs: boto3 "logs" service client, or None to create new
        :type logs: boto3.client
        :param cw: boto3 "cloudwatch" Service Resource, or None to create new
        :type cw: boto3.resource
        """
        self._func_name = func_name
        if logs is None:
            # override default max_attempts from 5 to 10, for throttling
            retry_conf = Config(
                retries={'max_attempts': BOTOCORE_MAX_ATTEMPTS})
            self._logs = boto3.client(
                'logs', config=retry_conf, region_name=region_name
            )
        else:
            self._logs = logs
        if cw is None:
            self._cw = boto3.resource('cloudwatch', region_name=region_name)
        else:
            self._cw = cw

    def get_filtered_logs(
            self, request_ids, interval=86400, group_name=None,
            always_match_re=None, never_match_re=None
    ):
        """
        Get CloudWatch logs for the last ``interval`` seconds and return only
        those entries with messages matching ``filter_re``.

        :param request_ids: list of str request IDs to get logs for
        :type request_ids: list
        :param group_name: CloudWatch logs group name. If left at default of
          ``None``, defaults to ``/aws/lambda/{func_name}``.
        :type group_name: str
        :param interval: how far back in logs to look, in seconds
        :type interval: int
        :param always_match_re: Regex for logs to ALWAYS return
        :type always_match_re: ``re``
        :param never_match_re: Regex for logs to NEVER return, even if they
          match ``always_match_re``.
        :type never_match_re: ``re``
        :return: dict of request_id to list of log entry dicts
        :rtype: dict
        """
        logs = self.get_cloudwatch_logs(
            interval=interval, group_name=group_name
        )
        if group_name is None:
            group_name = '/aws/lambda/%s' % self._func_name
        result = {}
        matchcount = 0
        for log in logs:
            m = self.req_id_re.match(log['message'])
            if always_match_re is None:
                always_m = None
            else:
                # if the regex is found, then skip this log and continue
                noelw = re.match(self.NO_OWNER_EMAIL_LOOKUP_WARNING,
                                 log['message'])
                if noelw:
                    continue
                always_m = always_match_re.match(log['message'])
            if m is None and always_m is None:
                logger.debug(
                    'Event %s in group %s stream %s does not match '
                    'RequestId regex: %s', log['eventId'], group_name,
                    log['logStreamName'], log['message']
                )
                continue
            if m is not None:
                req_id = m.group(2)
                if req_id in request_ids:
                    log['logGroupName'] = group_name
                    if req_id not in result:
                        result[req_id] = []
                    result[req_id].append(log)
                    matchcount += 1
            if always_m is not None:
                if (
                    never_match_re is not None and
                    never_match_re.match(log['message'])
                ):
                    logger.debug(
                        'Message matched always_match_re but also '
                        'never_match_re; suppressing: %s', log['message']
                    )
                    continue
                if 'always_match' not in result:
                    result['always_match'] = []
                result['always_match'].append(log)
        logger.debug(
            'Filtered %d log messages to %d messages from %d invocations',
            len(logs), matchcount, len(result)
        )
        return result

    def get_cloudwatch_logs(self, interval=86400, group_name=None):
        """
        Get CloudWatch logs for the last ``interval`` seconds. The log group
        name defaults to ``/aws/lambda/{func_name}`` if left at the default of
        None.

        :param group_name: CloudWatch logs group name. If left at default of
          ``None``, defaults to ``/aws/lambda/{func_name}``.
        :type group_name: str
        :param interval: how far back in logs to look, in seconds
        :type interval: int
        :return: list of log entry dicts, sorted by timestamp
        :rtype: list
        """
        interval = interval * 1000  # milliseconds
        now = int(time()) * 1000
        cutoff = now - interval
        if group_name is None:
            group_name = '/aws/lambda/%s' % self._func_name
        logger.debug('Finding streams in CW Log Group: %s', group_name)
        paginator = self._logs.get_paginator('describe_log_streams')
        stream_iterator = paginator.paginate(
            logGroupName=group_name,
            orderBy='LastEventTime',
            descending=True
        )
        streams = []
        try:
            for resp in stream_iterator:
                for stream in resp['logStreams']:
                    if stream.get('lastEventTimestamp', 0) < cutoff:
                        continue
                    streams.append(stream['logStreamName'])
        except Exception as ex:
            if hasattr(ex, 'response'):
                emsg = ex.response.get('Error', {}).get('Code', 'unknown')
                if emsg == 'ResourceNotFoundException':
                    logger.warning('CloudWatch Log group does not exist: %s',
                                   group_name)
                    return []
            raise
        logger.debug('Found %d log streams with events in time span',
                     len(streams))
        logs = []
        for sname in streams:
            logs.extend(self._get_cw_log_stream(
                group_name,
                sname,
                cutoff,
                now
            ))
        return sorted(logs, key=itemgetter('timestamp'))

    def _get_cw_log_stream(self, group_name, stream_name, start_ts, end_ts):
        """
        Return all log messages from the specified stream at or after ``ts``.

        :param group_name: CloudWatch log group name
        :type group_name: str
        :param stream_name: CloudWatch log stream name
        :type stream_name: str
        :param start_ts: timestamp in milliseconds to return logs after
        :type start_ts: int
        :param end_ts: timestamp in milliseconds to return logs before
        :type end_ts: int
        :return:
        :rtype: list
        """
        messages = []
        logger.debug('Getting events from CloudWatch Logs Group %s stream %s',
                     group_name, stream_name)
        paginator = self._logs.get_paginator('filter_log_events')
        resp_iter = paginator.paginate(
            logGroupName=group_name,
            logStreamNames=[stream_name],
            startTime=start_ts,
            endTime=end_ts
        )
        for resp in resp_iter:
            messages.extend(resp['events'])
        logger.debug('Found %d messages in stream %s',
                     len(messages), stream_name)
        return messages

    def get_cloudwatch_metric_sums(self, interval=86400, period=86400):
        """
        Return a dict of CloudWatch Metrics for this Lambda function, summed
        over ``interval``. Keys are metric names ("Errors", "Throttles",
        "Invocations") and values are sums of each ``period``-period datapoint,
        for the past ``interval`` seconds.

        For further information on these metrics, see:
        https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/
        lam-metricscollected.html

        :param interval: how many seconds of historical data to request
        :type interval: int
        :param period: the metric collection period to request from CloudWatch
        :type period: int
        :return: dict of metric name to sum for the last ``interval`` seconds
        :rtype: dict
        """
        res = {'Errors': 0.0, 'Throttles': 0.0, 'Invocations': 0.0}
        now = datetime.utcnow().replace(tzinfo=UTC())
        start = now - timedelta(seconds=interval)
        logger.debug('Checking CloudWatch Metrics for Lambda function: %s',
                     self._func_name)
        for m in self._cw.metrics.filter(
            Namespace='AWS/Lambda',
            Dimensions=[
                {
                    'Name': 'FunctionName',
                    'Value': self._func_name
                }
            ]
        ):
            dims = _name_value_dict(m.dimensions)
            # Lambda metrics are published twice, once with just
            # FunctionName, and a second time with both FunctionName and
            # Resource. Skip the duplicates that also have Resource dimension.
            if dims.keys() != ['FunctionName']:
                continue
            stats = m.get_statistics(
                Dimensions=[{'Name': 'FunctionName', 'Value': self._func_name}],
                StartTime=start,
                EndTime=now,
                Period=period,
                Statistics=['Sum']
            )
            val = sum([x['Sum'] for x in stats['Datapoints']])
            res[m.metric_name] = val
        logger.debug('Metrics for %s: %s', self._func_name, res)
        return res

    @staticmethod
    def find_matching_func_names(filter, region_name, client=None):
        """
        Return a list of all Lambda functions with names that either start
        with ``filter`` (if ``filter`` is a string) or match ``filter`` (if
        filter is a ``re.RegexObject``).

        :param filter: lambda function name filter
        :type filter: ``str`` ``re.RegexObject``
        :param region_name: region name to run against
        :type region_name: str
        :param client: boto3 Lambda client, or None to create new
        :type client: ``boto3.client``
        :return: list of matching Lambda function names
        :rtype: list
        """
        if client is None:
            client = boto3.client('lambda', region_name=region_name)
        if isinstance(filter, type('')):
            filter = re.compile('^' + re.escape(filter) + '.*')
        logger.debug(
            'Finding Lambda function names matching: %s', filter.pattern
        )
        matches = []
        total = 0
        paginator = client.get_paginator('list_functions')
        for response in paginator.paginate():
            for func in response['Functions']:
                total += 1
                if not filter.match(func['FunctionName']):
                    continue
                matches.append(func['FunctionName'])
        logger.debug('Matched %d of %d Lambda functions', len(matches), total)
        return sorted(matches)


class CustodianErrorReporter(object):
    """Scan and report on CW Metrics/Logs errors for c7n lambdas"""

    #: How far to look back in logs and metrics, in seconds
    INTERVAL = 86400

    #: Human-readable description of the interval
    INVL_DESC = 'day'

    #: What period to request CloudWatch metrics for
    METRIC_PERIOD = 86400

    #: Amount of time (float seconds) to sleep between checking each function,
    #: to try to avoid API rate limiting.
    INTER_FUNC_SLEEP = 3.0

    ALL_ERROR_FUNCTIONS = re.compile(r'^cloud-custodian.*')
    ALL_ERROR_LOG_RE = re.compile(r'.*(ERROR|WARNING).*')

    def __init__(self, config, region_name):
        """
        :param config: a non-region-specific config for this account
        :type config: CaisConfig
        :param region_name: the name of the region to run against
        :type region_name: str
        """
        self._config = config.config_for_region(region_name)
        self._region_name = region_name
        # override default max_attempts from 5 to 10, for throttling
        retry_conf = Config(retries={'max_attempts': BOTOCORE_MAX_ATTEMPTS})
        self._logs = boto3.client(
            'logs', config=retry_conf, region_name=region_name
        )
        self._cw = boto3.resource('cloudwatch', region_name=region_name)
        self._lambda = boto3.client('lambda', region_name=region_name)
        self._sqs = boto3.client('sqs', region_name=region_name)
        self._dlq_url = self._sqs_arn_to_url(
            self._config.dead_letter_queue_arn
        )
        self._now = datetime.now()
        self._start = self._now - timedelta(seconds=self.INTERVAL)
        self._failed_request_ids = {}  # set by _get_sqs_dlq()
        self._sqs_rcpts = []  # set by _get_sqs_dlq()

    def _sqs_arn_to_url(self, arn):
        """
        Find the URL for an SQS Queue given its ARN.

        :param arn: the ARN of the queue
        :type arn: str
        :return: the URL of the queue
        :rtype: str
        """
        m = re.match(r'^arn:aws:sqs:[^:]+:(\d+):([^:]+)$', arn)
        assert m is not None
        acct_id = m.group(1)
        queue_name = m.group(2)
        logger.debug(
            'Looking up SQS Queue URL for ARN %s (owner=%s name=%s)',
            arn, acct_id, queue_name
        )
        try:
            url = self._sqs.get_queue_url(
                QueueName=queue_name, QueueOwnerAWSAccountId=acct_id
            )['QueueUrl']
            logger.info('Found SQS Queue URL as %s for %s', url, arn)
            return url
        except self._sqs.exceptions.QueueDoesNotExist:
            logger.info('SQS Queue %s does not exist', arn)
            return None

    def run(self, never_match_re=None):
        """ collect and report on all cloud-custodian Lambda errors """
        print(
            'Searching cloud-custodian Lambda functions for failed invocations'
        )
        lambda_names = LambdaHealthChecker.find_matching_func_names(
            re.compile(r'^(custodian-|cloud-custodian-).*'), self._region_name
        )
        logger.debug('Custodian Lambda functions: %s', lambda_names)
        errors = False
        self._get_sqs_dlq()
        logger.debug(
            '%d failed Lambda invocations: %s',
            len(self._failed_request_ids), self._failed_request_ids.keys()
        )
        for fname in lambda_names:
            if not self._check_function(fname, never_match_re=never_match_re):
                logger.info(
                    '_check_function returned False (NOT HEALTHY) for: %s',
                    fname
                )
                errors = True
            logger.debug(
                'Sleeping %s seconds before checking next function',
                self.INTER_FUNC_SLEEP
            )
            sleep(self.INTER_FUNC_SLEEP)
        self._ack_sqs()
        req_ids = [
            i for i in self._failed_request_ids
            if self._failed_request_ids[i] is None
        ]
        if len(req_ids) > 0:
            print(
                "\n\n" +
                red('ERROR: %d failed Lambda RequestIDs could not be tied '
                    'to their function names: %s' % (len(req_ids), req_ids)) +
                "\n\n"
            )
        if errors:
            print('Some lambda functions had errors in the last '
                  '%s' % self.INVL_DESC)
            raise SystemExit(1)
        print('No Lambda functions had errors in the last ' + self.INVL_DESC)

    def _get_sqs_dlq(self):
        """
        Pull all messages from the SQS Dead Letter Queue. Add the failed Lambda
        RequestIDs to `self._failed_request_ids` and the SQS Reciept Handles
        to `self._sqs_rcpts`.
        """
        if self._dlq_url is None:
            logger.warning('Dead-letter SQS queue could not be found; skipping')
            return
        count = 0
        msgs = [None]
        logger.info('Polling SQS queue: %s', self._dlq_url)
        while len(msgs) > 0:
            response = self._sqs.receive_message(
                QueueUrl=self._dlq_url,
                WaitTimeSeconds=20,
                MaxNumberOfMessages=10,
                MessageAttributeNames=['RequestID', 'ErrorMessage'],
                AttributeNames=['SentTimestamp']
            )
            msgs = response.get('Messages', [])
            logger.debug('%d SQS Messages received from one poll', len(msgs))
            for m in msgs:
                count += 1
                self._failed_request_ids[
                    m['MessageAttributes']['RequestID']['StringValue']] = None
                self._sqs_rcpts.append(m['ReceiptHandle'])
        logger.info('Received %d SQS messages in total', count)
        logger.debug('SQS Message Receipt Handles: %s', self._sqs_rcpts)

    def _ack_sqs(self):
        """
        Delete (ack) all SQS messages in `self._sqs_rcpts`.
        """
        for rh in self._sqs_rcpts:
            self._sqs.delete_message(
                QueueUrl=self._dlq_url,
                ReceiptHandle=rh
            )

    def _check_function(self, func_name, never_match_re=None):
        """
        Check health of one Lambda function. Print information on it to STDOUT.
        Return True for healthy, False if errors/failures.

        :param func_name: Lambda function name to check
        :type func_name: str
        :param never_match_re: Regex for logs to NEVER return, even if they
          match ``always_match_re``.
        :type never_match_re: ``re``
        :return: whether the function had errors/failures
        :rtype: bool
        """
        c = LambdaHealthChecker(
            func_name, self._region_name, logs=self._logs, cw=self._cw
        )
        req_ids = [
            i for i in self._failed_request_ids
            if self._failed_request_ids[i] is None
        ]
        if self.ALL_ERROR_FUNCTIONS.match(func_name):
            logs = c.get_filtered_logs(
                req_ids, always_match_re=self.ALL_ERROR_LOG_RE,
                never_match_re=never_match_re
            )
        else:
            logs = c.get_filtered_logs(req_ids)
        metrics = c.get_cloudwatch_metric_sums()
        msg = []
        if metrics['Invocations'] > 0:
            throttle_pct = (metrics['Throttles'] / metrics['Invocations']) * 100
            error_pct = (metrics['Errors'] / metrics['Invocations']) * 100
        else:
            throttle_pct = 0
            error_pct = 0
        if error_pct > 50:
            msg.append('Lambda Function Errors: %s%% (%d of %d invocations)' % (
                error_pct,
                metrics['Errors'],
                metrics['Invocations']
            ))
        if throttle_pct > 50:
            msg.append(
                'Lambda Function Throttles: %s%% (%d of %d invocations)' % (
                    throttle_pct,
                    metrics['Throttles'],
                    metrics['Invocations']
                )
            )
        if len(logs) < 1 and len(msg) == 0:
            print(green('%s: OK\n' % func_name))
            return True
        print(red('%s: ERRORS' % func_name))
        for m in msg:
            print("\t%s" % red(m))
        if len(logs) < 1:
            return True
        print("\n\tLogs For Failed Invocations:\n")
        for req_id in logs.keys():
            if req_id == 'always_match':
                continue
            events = logs[req_id]
            self._failed_request_ids[req_id] = func_name
            print("\t" + red('RequestID=%s logGroupName=%s logStreamName=%s' % (
                req_id, events[0]['logGroupName'], events[0]['logStreamName']
            )))
            for e in events:
                print("\n".join([
                    "\t\t%s" % line.replace("\t", ' ')
                    for line in e['message'].split("\n")
                    if line.strip() != ''
                ]))
        if 'always_match' in logs:
            print("\t" + red(
                'Always-Match Logs (RequestID not in DLQ, but log matches '
                'regex that we want to always alarm on)'
            ))
            for e in logs['always_match']:
                print("\n".join([
                    "\t\t%s" % line.replace("\t", ' ')
                    for line in e['message'].split("\n")
                    if line.strip() != ''
                ]))
        print('')
        return False


def _name_value_dict(lst):
    """
    Given a list (``lst``) containing dicts with ``Name`` and ``Value`` keys,
    return a single dict of Name -> Value.
    """
    res = {}
    for item in lst:
        res[item['Name']] = item['Value']
    return res


def parse_args(argv):
    p = argparse.ArgumentParser(
        description='Report on c7n lambda errors',
        epilog='This tool is part of manheim_c7n_tools v%s.\n'
               'For documentation, see: %s' % (VERSION, PROJECT_URL),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    p.add_argument('-V', '--version', action='version', version=VERSION)
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-A', '--no-assume-role', dest='assume_role',
                   action='store_false', default=True,
                   help='Do not assume a role, even if  specified in the '
                        'configuration file.')
    p.add_argument('-c', '--config', dest='config', action='store',
                   default='manheim-c7n-tools.yml',
                   help='Config file path (default: ./manheim-c7n-tools.yml)')
    p.add_argument('-n', '--never-match-re', dest='never_match_re', type=str,
                   action='store', default=None,
                   help='Regex for Lambda function logs to suppress/never '
                        'match')
    p.add_argument('ACCOUNT_NAME', action='store', type=str,
                   help='Account name to run errorscan against')
    p.add_argument('REGION_NAME', action='store', type=str,
                   help='AWS Region name to run errorscan against')
    args = p.parse_args(argv)
    return args


def main():
    args = parse_args(sys.argv[1:])

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    conf = ManheimConfig.from_file(args.config, args.ACCOUNT_NAME)
    if args.assume_role:
        assume_role(conf)
    if args.never_match_re is not None:
        args.never_match_re = re.compile(args.never_match_re)
    CustodianErrorReporter(conf, args.REGION_NAME).run(
        never_match_re=args.never_match_re
    )


if __name__ == "__main__":
    main()



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import subprocess
import re
import os

import boto3

logger = logging.getLogger(__name__)


def set_log_info(log):
    """
    set log level to INFO with appropriate format

    :param log: the logger to set level and format on
    :type log: logging.Logger
    """
    set_log_level_format(log, logging.INFO,
                         '%(asctime)s %(levelname)s:%(name)s:%(message)s')


def set_log_debug(log):
    """
    set log level to DEBUG, and debug-level output format

    :param log: the logger to set level and format on
    :type log: logging.Logger
    """
    set_log_level_format(
        log,
        logging.DEBUG,
        "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - "
        "%(name)s.%(funcName)s() ] %(message)s"
    )


def set_log_level_format(log, level, format):
    """
    Set log level and format.

    :param log: the logger to set level and format on
    :type log: logging.Logger
    :param level: logging level; see the :py:mod:`logging` constants.
    :type level: int
    :param format: logging formatter format string
    :type format: str
    """
    formatter = logging.Formatter(fmt=format)
    log.handlers[0].setFormatter(formatter)
    log.setLevel(level)


def red(s):
    """
    Return the given string (``s``) surrounded by the ANSI escape codes to
    print it in red.
    :param s: string to console-color red
    :type s: str
    :returns: s surrounded by ANSI color escapes for red text
    :rtype: str
    """
    return "\033[0;31m" + s + "\033[0m"


def green(s):
    """
    Return the given string (``s``) surrounded by the ANSI escape codes to
    print it in green.
    :param s: string to console-color green
    :type s: str
    :returns: s surrounded by ANSI color escapes for green text
    :rtype: str
    """
    return "\033[0;32m" + s + "\033[0m"


def bold(s):
    """
    Return the given string (``s``) surrounded by the ANSI escape codes to
    print it in bold.
    :param s: string to console format as bold
    :type s: str
    :returns: s surrounded by ANSI color escapes for green text
    :rtype: str
    """
    return "\033[1m" + s + "\033[0m"


def git_html_url():
    """
    Run ``git config remote.origin.url`` in the current directory. Assuming it
    works, return the HTML URL for the repository (assumes github.com or
    github enterprise).

    :return: repository HTML base URL
    :rtype: str
    :raises: RuntimeError if the command fails or the URL cannot be parsed
    """
    p = subprocess.check_output(
        ['git', 'config', 'remote.origin.url'], text=True
    )
    p = p.strip()

    # git / ssh remote
    m = re.match(
        r'^([^@]+)@(?P<hostname>[a-zA-Z0-9_-]+(\.[a-zA-Z0-9_-]+)+):'
        r'(?P<path>.*/[^\.]*)(\.git)?$',
        p
    )
    if not m:
        m = re.match(
            r'^https?://(?P<hostname>[a-zA-Z0-9_-]+(\.[a-zA-Z0-9_-]+)+)/'
            r'(?P<path>.*/[^\.]*)(\.git)?$',
            p
        )
    if not m:
        raise RuntimeError(
            'Error: no pattern match for URL: %s' % p
        )
    return 'https://%s/%s/' % (m.group('hostname'), m.group('path'))


def assume_role(config):
    """
    Call sts:AssumeRole (via boto3) to assume the role specified by the
    configuration. Export the resulting credentials as environment variables
    for the current process.

    The configuration is taken from the ``assume_role`` section of the config.

    :param config: ManheimConfig object containing assume_role configuration
    :type config: ManheimConfig
    """
    try:
        conf = config.assume_role
    except AttributeError:
        logger.debug('No assume_role configuration; not assuming a role.')
        return
    kwargs = {
        'RoleArn': conf['role_arn'],
        'RoleSessionName': 'manheim-c7n-tools_%s' % config.account_name
    }
    if 'external_id' in conf:
        kwargs['ExternalId'] = conf['external_id']
    if 'duration_seconds' in conf:
        kwargs['DurationSeconds'] = int(conf['duration_seconds'])
    logger.info(
        'Calling sts:AssumeRole via boto3 with arguments: %s', kwargs
    )
    # We need to prevent STS from using the botocore/boto3 default session,
    # or else the default session will have the creds from the previous
    # account, not the assumed role, and none of this will work...
    sess = boto3.session.Session(region_name='us-east-1')
    sts = sess.client('sts')
    resp = sts.assume_role(**kwargs)
    os.environ['AWS_ACCESS_KEY_ID'] = resp['Credentials']['AccessKeyId']
    os.environ['AWS_SECRET_ACCESS_KEY'] = resp[
        'Credentials'
    ]['SecretAccessKey']
    os.environ['AWS_SESSION_TOKEN'] = resp['Credentials']['SessionToken']
    logger.info(
        'Exported AssumeRole credentials; AccessKeyId %s expires at %s; '
        'AssumedRoleUser ARN: %s', resp['Credentials']['AccessKeyId'],
        resp['Credentials']['Expiration'],
        resp['AssumedRoleUser']['Arn']
    )



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Script to clean up the custodian S3 buckets, by moving logs from any deleted
policies to an "archived-logs/" prefix.
"""

import sys
import logging
import boto3
import argparse

import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from manheim_c7n_tools.utils import set_log_info, set_log_debug
from manheim_c7n_tools.version import VERSION

logger = logging.getLogger(__name__)


class S3Archiver(object):

    def __init__(self, region_name, bucket_name, conf_file, dryrun=False):
        logger.info('Connecting to S3 in %s for bucket %s (config file: %s)',
                    region_name, bucket_name, conf_file)
        self._s3 = boto3.resource('s3', region_name=region_name)
        self._region_name = region_name
        self._bucket_name = bucket_name
        self._bucket = self._s3.Bucket(bucket_name)
        self._conf_file = conf_file
        self._dryrun = dryrun

    def run(self):
        policy_names = self._get_policy_names()
        logger.debug('Found %d policies in config', len(policy_names))
        prefixes = self._get_s3_policy_prefixes()
        logger.debug('Found %d policy prefixes in S3 bucket', len(prefixes))
        move_count = 0
        for p in prefixes:
            if p not in policy_names:
                self._move_prefix(p)
                move_count += 1
        logger.info('Archived %d policy name prefixes', move_count)

    def _move_prefix(self, policy_name):
        """
        Given a policy name key prefix in S3, move everything under it to the
        ``archived-logs/`` prefix.

        :param policy_name: name of the policy prefix in S3
        :type policy_name: str
        """
        if not self._dryrun:
            logger.info('Moving policy prefix logs/%s to archived-logs/%s',
                        policy_name, policy_name)
        count = 0
        for o in self._bucket.objects.filter(
            Prefix='logs/%s/' % policy_name
        ):
            dest = o.key.replace('logs/', 'archived-logs/')
            self._s3_move_file(o, dest)
            count += 1
        if self._dryrun:
            logger.info('DRYRUN: Would move %d objects under logs/%s to '
                        'archived-logs/%s', count, policy_name, policy_name)
        else:
            logger.info('Moved %d objects under logs/%s to archived-logs/%s',
                        count, policy_name, policy_name)

    def _s3_move_file(self, obj_summary, dest_key):
        """
        S3 doesn't have any built-in logic for "moving" or "renaming" an object.
        The way `awscli` and all of the SDK examples do this is by copying the
        source to the destination, then deleting the source. To make this a bit
        more unweildy, it's far easier to copy with the boto3 client than with
        the fancy Resource-oriented API.

        :param obj_summary: the S3 ObjectSummary instance to move
        :type obj_summary: ``boto3.S3.ObjectSummary``
        :param dest_key: S3 key to move to
        :type dest_key: str
        """
        if self._dryrun:
            logger.debug('DRYRUN: Would move %s to %s',
                         obj_summary.key, dest_key)
            return
        # ELSE not a dry run, actually do it
        client = self._bucket.meta.client
        logger.debug('Copying %s to %s', obj_summary.key, dest_key)
        client.copy_object(
            ACL='private',
            Bucket=self._bucket.name,
            Key=dest_key,
            CopySource={
                'Bucket': self._bucket.name,
                'Key': obj_summary.key
            },
            MetadataDirective='COPY',
            TaggingDirective='COPY'
        )
        # ok, copied, now delete
        logger.debug('Deleting %s', obj_summary.key)
        obj_summary.delete()

    def _get_s3_policy_prefixes(self):
        """
        Find all of the per-policy prefixes (a.k.a. "directories") in the S3
        bucket. Return a list of them

        :return: list of per-policy prefixes in S3 bucket
        :rtype: list
        """
        client = self._bucket.meta.client
        response = client.list_objects(
            Bucket=self._bucket.name,
            Delimiter='/',
            Prefix='logs/'
        )
        if response['IsTruncated']:
            raise RuntimeError('ERROR: S3 response was truncated!')
        result = []
        for pname in response.get('CommonPrefixes', []):
            result.append(pname['Prefix'].replace('logs/', '').strip('/'))
        return result

    def _get_policy_names(self):
        """
        Read the custodian config file; return a list of policy names.

        :return: list of policy names
        :rtype: list
        """
        with open(self._conf_file, 'r') as fh:
            contents = fh.read()
        data = yaml.load(contents, Loader=SafeLoader)
        return [p['name'] for p in data['policies']]


def parse_args(argv):
    p = argparse.ArgumentParser(
        description='Archive S3 logs for deleted policies'
    )
    p.add_argument('-V', '--version', action='version', version=VERSION)
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-d', '--dry-run', dest='dryrun', action='store_true',
                   default=False,
                   help='print what would be done; dont move anything')
    p.add_argument('REGION_NAME', action='store', type=str,
                   help='AWS region name to run against')
    p.add_argument('BUCKET_NAME', action='store', type=str,
                   help='S3 Bucket Name')
    p.add_argument('CONF_FILE', action='store', type=str,
                   help='path to cloud-custodian config YML file')
    args = p.parse_args(argv)
    return args


def main():
    # setup logging for direct command-line use
    global logger
    FORMAT = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.INFO, format=FORMAT)
    logger = logging.getLogger()

    # suppress boto3 internal logging below WARNING level
    boto3_log = logging.getLogger("boto3")
    boto3_log.setLevel(logging.WARNING)
    boto3_log.propagate = True

    # suppress botocore internal logging below WARNING level
    botocore_log = logging.getLogger("botocore")
    botocore_log.setLevel(logging.WARNING)
    botocore_log.propagate = True
    # end setup logging
    args = parse_args(sys.argv[1:])

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    S3Archiver(
        args.REGION_NAME, args.BUCKET_NAME, args.CONF_FILE, dryrun=args.dryrun
    ).run()


if __name__ == "__main__":
    main()



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pytest
import os
from datetime import datetime

from manheim_c7n_tools.utils import (
    set_log_debug, set_log_info, set_log_level_format, red, green, bold,
    git_html_url, assume_role
)
from manheim_c7n_tools.config import ManheimConfig

from mock import patch, call, Mock, PropertyMock  # noqa

pbm = 'manheim_c7n_tools.utils'


class TestUtils(object):

    def test_set_log_info(self):
        mock_log = Mock(spec_set=logging.Logger)
        with patch('%s.set_log_level_format' % pbm, autospec=True) as mock_set:
            set_log_info(mock_log)
        assert mock_set.mock_calls == [
            call(
                mock_log, logging.INFO,
                '%(asctime)s %(levelname)s:%(name)s:%(message)s'
            )
        ]

    def test_set_log_debug(self):
        mock_log = Mock(spec_set=logging.Logger)
        with patch('%s.set_log_level_format' % pbm, autospec=True) as mock_set:
            set_log_debug(mock_log)
        assert mock_set.mock_calls == [
            call(mock_log, logging.DEBUG,
                 "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - "
                 "%(name)s.%(funcName)s() ] %(message)s")
        ]

    def test_set_log_level_format(self):
        mock_log = Mock(spec_set=logging.Logger)
        mock_handler = Mock(spec_set=logging.Handler)
        type(mock_log).handlers = [mock_handler]
        with patch(
            '%s.logging.Formatter' % pbm, autospec=True
        ) as mock_formatter:
            set_log_level_format(mock_log, 5, 'foo')
        assert mock_formatter.mock_calls == [
            call(fmt='foo')
        ]
        assert mock_handler.mock_calls == [
            call.setFormatter(mock_formatter.return_value)
        ]
        assert mock_log.mock_calls == [
            call.setLevel(5)
        ]

    def test_red(self):
        assert red('foo') == "\033[0;31mfoo\033[0m"

    def test_green(self):
        assert green('foo') == "\033[0;32mfoo\033[0m"

    def test_bold(self):
        assert bold('foo') == "\033[1mfoo\033[0m"


class TestGitHtmlUrl(object):

    def test_private_git(self):
        with patch(
            '%s.subprocess.check_output' % pbm, autospec=True
        ) as mock_co:
            mock_co.return_value = 'git@git.example.com:Foo/bar.git'
            res = git_html_url()
        assert res == 'https://git.example.com/Foo/bar/'
        assert mock_co.mock_calls == [
            call(['git', 'config', 'remote.origin.url'], text=True)
        ]

    def test_private_https(self):
        with patch(
            '%s.subprocess.check_output' % pbm, autospec=True
        ) as mock_co:
            mock_co.return_value = 'https://git.example.com/Foo/bar.git'
            res = git_html_url()
        assert res == 'https://git.example.com/Foo/bar/'
        assert mock_co.mock_calls == [
            call(['git', 'config', 'remote.origin.url'], text=True)
        ]

    def test_github_git(self):
        with patch(
            '%s.subprocess.check_output' % pbm, autospec=True
        ) as mock_co:
            mock_co.return_value = 'git@github.com:Foo/bar.git'
            res = git_html_url()
        assert res == 'https://github.com/Foo/bar/'
        assert mock_co.mock_calls == [
            call(['git', 'config', 'remote.origin.url'], text=True)
        ]

    def test_github_https(self):
        with patch(
            '%s.subprocess.check_output' % pbm, autospec=True
        ) as mock_co:
            mock_co.return_value = 'https://github.com/Foo/bar.git'
            res = git_html_url()
        assert res == 'https://github.com/Foo/bar/'
        assert mock_co.mock_calls == [
            call(['git', 'config', 'remote.origin.url'], text=True)
        ]

    def test_bad_pattern(self):
        with patch(
            '%s.subprocess.check_output' % pbm, autospec=True
        ) as mock_co:
            with pytest.raises(RuntimeError):
                mock_co.return_value = 'foobar'
                git_html_url()


class TestAssumeRole(object):

    def setup(self):
        self.m_conf = Mock(spec_set=ManheimConfig)
        type(self.m_conf).account_name = PropertyMock(return_value='aName')

    def test_success(self):
        m_sts = Mock()
        m_sts.assume_role.return_value = {
            'Credentials': {
                'AccessKeyId': 'AKID',
                'SecretAccessKey': 'SKey',
                'SessionToken': 'SToken',
                'Expiration': datetime(2018, 10, 8, 12, 13, 14)
            },
            'AssumedRoleUser': {
                'AssumedRoleId': 'ARid',
                'Arn': 'UserARN'
            },
            'PackedPolicySize': 123
        }
        m_sess = Mock()
        m_sess.client.return_value = m_sts
        type(self.m_conf).assume_role = PropertyMock(return_value={
            'role_arn': 'assumeRoleArn'
        })
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch.dict(os.environ, {}, clear=True):
                with patch('%s.boto3.session.Session' % pbm) as mock_boto:
                    mock_boto.return_value = m_sess
                    assume_role(self.m_conf)
                    assert os.environ == {
                        'AWS_ACCESS_KEY_ID': 'AKID',
                        'AWS_SECRET_ACCESS_KEY': 'SKey',
                        'AWS_SESSION_TOKEN': 'SToken'
                    }
        expected_args = {
            'RoleArn': 'assumeRoleArn',
            'RoleSessionName': 'manheim-c7n-tools_aName'
        }
        assert mock_boto.mock_calls == [
            call(region_name='us-east-1'),
            call().client('sts'),
            call().client().assume_role(**expected_args)
        ]
        assert mock_logger.mock_calls == [
            call.info(
                'Calling sts:AssumeRole via boto3 with arguments: %s',
                expected_args
            ),
            call.info(
                'Exported AssumeRole credentials; AccessKeyId %s expires at '
                '%s; AssumedRoleUser ARN: %s', 'AKID',
                datetime(2018, 10, 8, 12, 13, 14), 'UserARN'
            )
        ]

    def test_success_all_options(self):
        m_sts = Mock()
        m_sts.assume_role.return_value = {
            'Credentials': {
                'AccessKeyId': 'AKID',
                'SecretAccessKey': 'SKey',
                'SessionToken': 'SToken',
                'Expiration': datetime(2018, 10, 8, 12, 13, 14)
            },
            'AssumedRoleUser': {
                'AssumedRoleId': 'ARid',
                'Arn': 'UserARN'
            },
            'PackedPolicySize': 123
        }
        m_sess = Mock()
        m_sess.client.return_value = m_sts
        type(self.m_conf).assume_role = PropertyMock(return_value={
            'role_arn': 'assumeRoleArn',
            'external_id': 'eID',
            'duration_seconds': '1234'
        })
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch.dict(os.environ, {}, clear=True):
                with patch('%s.boto3.session.Session' % pbm) as mock_boto:
                    mock_boto.return_value = m_sess
                    assume_role(self.m_conf)
                    assert os.environ == {
                        'AWS_ACCESS_KEY_ID': 'AKID',
                        'AWS_SECRET_ACCESS_KEY': 'SKey',
                        'AWS_SESSION_TOKEN': 'SToken'
                    }
        expected_args = {
            'RoleArn': 'assumeRoleArn',
            'RoleSessionName': 'manheim-c7n-tools_aName',
            'ExternalId': 'eID',
            'DurationSeconds': 1234
        }
        assert mock_boto.mock_calls == [
            call(region_name='us-east-1'),
            call().client('sts'),
            call().client().assume_role(**expected_args)
        ]
        assert mock_logger.mock_calls == [
            call.info(
                'Calling sts:AssumeRole via boto3 with arguments: %s',
                expected_args
            ),
            call.info(
                'Exported AssumeRole credentials; AccessKeyId %s expires at '
                '%s; AssumedRoleUser ARN: %s', 'AKID',
                datetime(2018, 10, 8, 12, 13, 14), 'UserARN'
            )
        ]

    def test_no_role_arn(self):
        m_sts = Mock()
        m_sts.assume_role.return_value = {
            'Credentials': {
                'AccessKeyId': 'AKID',
                'SecretAccessKey': 'SKey',
                'SessionToken': 'SToken',
                'Expiration': datetime(2018, 10, 8, 12, 13, 14)
            },
            'AssumedRoleUser': {
                'AssumedRoleId': 'ARid',
                'Arn': 'UserARN'
            },
            'PackedPolicySize': 123
        }
        m_sess = Mock()
        m_sess.client.return_value = m_sts
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch.dict(os.environ, {}, clear=True):
                with patch('%s.boto3.session.Session' % pbm) as mock_boto:
                    mock_boto.return_value = m_sess
                    assume_role(self.m_conf)
                    assert os.environ == {}
        assert mock_boto.mock_calls == []
        assert mock_logger.mock_calls == [
            call.debug('No assume_role configuration; not assuming a role.')
        ]
